@@ -14,10 +14,11 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
-import futurist
 import testtools
+import time
 
 import taskflow.engines
+from taskflow.engines.action_engine import executor
 from taskflow import exceptions as exc
 from taskflow.patterns import graph_flow as gf
 from taskflow.patterns import linear_flow as lf
@@ -104,7 +105,7 @@ class RetryTest(utils.EngineTestBase):
         engine = self._make_engine(flow)
         engine.storage.inject({'y': 4})
         with utils.CaptureListener(engine) as capturer:
-            self.assertRaisesRegexp(RuntimeError, '^Woot', engine.run)
+            self.assertRaisesRegex(RuntimeError, '^Woot', engine.run)
         self.assertEqual({'y': 4}, engine.storage.fetch_all())
         expected = ['flow-1.f RUNNING',
                     'r1.r RUNNING',
@@ -143,7 +144,7 @@ class RetryTest(utils.EngineTestBase):
         engine = self._make_engine(flow)
         engine.storage.inject({'y': 4})
         with utils.CaptureListener(engine) as capturer:
-            self.assertRaisesRegexp(RuntimeError, '^Gotcha', engine.run)
+            self.assertRaisesRegex(RuntimeError, '^Gotcha', engine.run)
         self.assertEqual({'y': 4, 'x': 1}, engine.storage.fetch_all())
         expected = ['flow-1.f RUNNING',
                     'r1.r RUNNING',
@@ -356,7 +357,7 @@ class RetryTest(utils.EngineTestBase):
                     'task1.t SUCCESS(5)',
                     'task2.t SUCCESS(None)',
                     'flow-1.f SUCCESS']
-        self.assertItemsEqual(capturer.values, expected)
+        self.assertCountEqual(capturer.values, expected)
 
     def test_nested_flow_reverts_parent_retries(self):
         retry1 = retry.Times(3, 'r1', provides='x')
@@ -469,7 +470,7 @@ class RetryTest(utils.EngineTestBase):
         engine = self._make_engine(flow)
         engine.storage.inject({'y': 2})
         with utils.CaptureListener(engine) as capturer:
-            self.assertRaisesRegexp(RuntimeError, '^Woot', engine.run)
+            self.assertRaisesRegex(RuntimeError, '^Woot', engine.run)
         self.assertEqual({'y': 2}, engine.storage.fetch_all())
         expected = ['flow-1.f RUNNING',
                     'r1.r RUNNING',
@@ -495,8 +496,58 @@ class RetryTest(utils.EngineTestBase):
         flow = lf.Flow('test', retry=utils.OneReturnRetry(provides='x')).add(
             utils.FailingTask('fail'))
         engine = self._make_engine(flow)
-        self.assertRaisesRegexp(RuntimeError, '^Woot', engine.run)
-        self.assertRaisesRegexp(RuntimeError, '^Woot', engine.run)
+        self.assertRaisesRegex(RuntimeError, '^Woot', engine.run)
+        self.assertRaisesRegex(RuntimeError, '^Woot', engine.run)
+
+    def test_restart_reverted_unordered_flows_with_retries(self):
+        now = time.time()
+
+        # First flow of an unordered flow:
+        subflow1 = lf.Flow('subflow1')
+
+        # * a task that completes in 3 sec with a few retries
+        subsubflow1 = lf.Flow('subflow1.subsubflow1',
+                              retry=utils.RetryFiveTimes())
+        subsubflow1.add(utils.SuccessAfter3Sec('subflow1.fail1',
+                                               inject={'start_time': now}))
+        subflow1.add(subsubflow1)
+
+        # * a task that fails and triggers a revert after 5 retries
+        subsubflow2 = lf.Flow('subflow1.subsubflow2',
+                              retry=utils.RetryFiveTimes())
+        subsubflow2.add(utils.FailingTask('subflow1.fail2'))
+        subflow1.add(subsubflow2)
+
+        # Second flow of the unordered flow:
+        subflow2 = lf.Flow('subflow2')
+
+        # * a task that always fails and retries
+        subsubflow1 = lf.Flow('subflow2.subsubflow1',
+                              retry=utils.AlwaysRetry())
+        subsubflow1.add(utils.FailingTask('subflow2.fail1'))
+        subflow2.add(subsubflow1)
+
+        unordered_flow = uf.Flow('unordered_flow')
+        unordered_flow.add(subflow1, subflow2)
+
+        # Main flow, contains a simple task and an unordered flow
+        flow = lf.Flow('test')
+        flow.add(utils.NoopTask('task1'))
+        flow.add(unordered_flow)
+
+        engine = self._make_engine(flow)
+
+        # This test fails when using Green threads, skipping it for now
+        if isinstance(engine._task_executor,
+                      executor.ParallelGreenThreadTaskExecutor):
+            self.skipTest("Skipping this test when using green threads.")
+
+        with utils.CaptureListener(engine) as capturer:
+            self.assertRaisesRegex(exc.WrappedFailure,
+                                   '.*RuntimeError: Woot!',
+                                   engine.run)
+            # task1 should have been reverted
+            self.assertIn('task1.t REVERTED(None)', capturer.values)
 
     def test_run_just_retry(self):
         flow = utils.OneReturnRetry(provides='x')
@@ -597,7 +648,7 @@ class RetryTest(utils.EngineTestBase):
             utils.FailingTask('t2'))
         engine = self._make_engine(flow)
         with utils.CaptureListener(engine) as capturer:
-            self.assertRaisesRegexp(RuntimeError, '^Woot', engine.run)
+            self.assertRaisesRegex(RuntimeError, '^Woot', engine.run)
         expected = ['flow-1.f RUNNING',
                     'r1.r RUNNING',
                     'r1.r SUCCESS(1)',
@@ -646,7 +697,7 @@ class RetryTest(utils.EngineTestBase):
         flow = lf.Flow('flow-1', retry1).add(utils.FailingTaskWithOneArg('t1'))
         engine = self._make_engine(flow)
         with utils.CaptureListener(engine) as capturer:
-            self.assertRaisesRegexp(RuntimeError, '^Woot', engine.run)
+            self.assertRaisesRegex(RuntimeError, '^Woot', engine.run)
         expected = ['flow-1.f RUNNING',
                     'r1.r RUNNING',
                     'r1.r SUCCESS(3)',
@@ -689,7 +740,7 @@ class RetryTest(utils.EngineTestBase):
         flow = lf.Flow('flow-1', retry1).add(utils.FailingTaskWithOneArg('t1'))
         engine = self._make_engine(flow)
         with utils.CaptureListener(engine) as capturer:
-            self.assertRaisesRegexp(RuntimeError, '^Woot', engine.run)
+            self.assertRaisesRegex(RuntimeError, '^Woot', engine.run)
         expected = ['flow-1.f RUNNING',
                     'r1.r RUNNING',
                     'r1.r SUCCESS(2)',
@@ -716,7 +767,7 @@ class RetryTest(utils.EngineTestBase):
                     'r1.r REVERTING',
                     'r1.r REVERTED(None)',
                     'flow-1.f REVERTED']
-        self.assertItemsEqual(capturer.values, expected)
+        self.assertCountEqual(capturer.values, expected)
 
     def test_nested_for_each_revert(self):
         collection = [3, 2, 3, 5]
@@ -729,7 +780,7 @@ class RetryTest(utils.EngineTestBase):
         )
         engine = self._make_engine(flow)
         with utils.CaptureListener(engine) as capturer:
-            self.assertRaisesRegexp(RuntimeError, '^Woot', engine.run)
+            self.assertRaisesRegex(RuntimeError, '^Woot', engine.run)
         expected = ['flow-1.f RUNNING',
                     'task1.t RUNNING',
                     'task1.t SUCCESS(5)',
@@ -779,7 +830,7 @@ class RetryTest(utils.EngineTestBase):
         )
         engine = self._make_engine(flow)
         with utils.CaptureListener(engine) as capturer:
-            self.assertRaisesRegexp(RuntimeError, '^Woot', engine.run)
+            self.assertRaisesRegex(RuntimeError, '^Woot', engine.run)
         expected = ['flow-1.f RUNNING',
                     'task1.t RUNNING',
                     'task1.t SUCCESS(5)',
@@ -826,7 +877,7 @@ class RetryTest(utils.EngineTestBase):
         flow = lf.Flow('flow-1', retry1).add(utils.ConditionalTask('t1'))
         engine = self._make_engine(flow)
         engine.storage.inject({'y': 1})
-        self.assertRaisesRegexp(exc.NotFound, '^No elements left', engine.run)
+        self.assertRaisesRegex(exc.NotFound, '^No elements left', engine.run)
 
     def test_parameterized_for_each_with_list(self):
         values = [3, 2, 5]
@@ -835,7 +886,7 @@ class RetryTest(utils.EngineTestBase):
         engine = self._make_engine(flow)
         engine.storage.inject({'values': values, 'y': 1})
         with utils.CaptureListener(engine) as capturer:
-            self.assertRaisesRegexp(RuntimeError, '^Woot', engine.run)
+            self.assertRaisesRegex(RuntimeError, '^Woot', engine.run)
         expected = ['flow-1.f RUNNING',
                     'r1.r RUNNING',
                     'r1.r SUCCESS(3)',
@@ -871,7 +922,7 @@ class RetryTest(utils.EngineTestBase):
         engine = self._make_engine(flow)
         engine.storage.inject({'values': values, 'y': 1})
         with utils.CaptureListener(engine) as capturer:
-            self.assertRaisesRegexp(RuntimeError, '^Woot', engine.run)
+            self.assertRaisesRegex(RuntimeError, '^Woot', engine.run)
         expected = ['flow-1.f RUNNING',
                     'r1.r RUNNING',
                     'r1.r SUCCESS(3)',
@@ -898,7 +949,7 @@ class RetryTest(utils.EngineTestBase):
                     'r1.r REVERTING',
                     'r1.r REVERTED(None)',
                     'flow-1.f REVERTED']
-        self.assertItemsEqual(capturer.values, expected)
+        self.assertCountEqual(capturer.values, expected)
 
     def test_nested_parameterized_for_each_revert(self):
         values = [3, 2, 5]
@@ -912,7 +963,7 @@ class RetryTest(utils.EngineTestBase):
         engine = self._make_engine(flow)
         engine.storage.inject({'values': values, 'y': 1})
         with utils.CaptureListener(engine) as capturer:
-            self.assertRaisesRegexp(RuntimeError, '^Woot', engine.run)
+            self.assertRaisesRegex(RuntimeError, '^Woot', engine.run)
         expected = ['flow-1.f RUNNING',
                     'task-1.t RUNNING',
                     'task-1.t SUCCESS(5)',
@@ -956,7 +1007,7 @@ class RetryTest(utils.EngineTestBase):
         engine = self._make_engine(flow)
         engine.storage.inject({'values': values, 'y': 1})
         with utils.CaptureListener(engine) as capturer:
-            self.assertRaisesRegexp(RuntimeError, '^Woot', engine.run)
+            self.assertRaisesRegex(RuntimeError, '^Woot', engine.run)
         expected = ['flow-1.f RUNNING',
                     'task-1.t RUNNING',
                     'task-1.t SUCCESS(5)',
@@ -995,7 +1046,7 @@ class RetryTest(utils.EngineTestBase):
         flow = lf.Flow('flow-1', retry1).add(utils.ConditionalTask('t1'))
         engine = self._make_engine(flow)
         engine.storage.inject({'values': values, 'y': 1})
-        self.assertRaisesRegexp(exc.NotFound, '^No elements left', engine.run)
+        self.assertRaisesRegex(exc.NotFound, '^No elements left', engine.run)
 
     def _pretend_to_run_a_flow_and_crash(self, when):
         flow = uf.Flow('flow-1', retry.Times(3, provides='x')).add(
@@ -1110,7 +1161,7 @@ class RetryTest(utils.EngineTestBase):
         r = FailingRetry()
         flow = lf.Flow('testflow', r)
         engine = self._make_engine(flow)
-        self.assertRaisesRegexp(ValueError, '^OMG', engine.run)
+        self.assertRaisesRegex(ValueError, '^OMG', engine.run)
         self.assertEqual(1, len(engine.storage.get_retry_histories()))
         self.assertEqual(0, len(r.history))
         self.assertEqual([], list(r.history.outcomes_iter()))
@@ -1121,7 +1172,7 @@ class RetryTest(utils.EngineTestBase):
         r = NastyFailingRetry()
         flow = lf.Flow('testflow', r)
         engine = self._make_engine(flow)
-        self.assertRaisesRegexp(ValueError, '^WOOT', engine.run)
+        self.assertRaisesRegex(ValueError, '^WOOT', engine.run)
 
     def test_nested_provides_graph_reverts_correctly(self):
         flow = gf.Flow("test").add(
@@ -1136,7 +1187,7 @@ class RetryTest(utils.EngineTestBase):
         engine.storage.save('b', 11)
         engine.storage.save('a', 10)
         with utils.CaptureListener(engine, capture_flow=False) as capturer:
-            self.assertRaisesRegexp(RuntimeError, '^Woot', engine.run)
+            self.assertRaisesRegex(RuntimeError, '^Woot', engine.run)
         expected = ['c.t RUNNING',
                     'c.t FAILURE(Failure: RuntimeError: Woot!)',
                     'a.t REVERTING',
@@ -1145,7 +1196,7 @@ class RetryTest(utils.EngineTestBase):
                     'c.t REVERTED(None)',
                     'b.t REVERTING',
                     'b.t REVERTED(None)']
-        self.assertItemsEqual(capturer.values[:8], expected)
+        self.assertCountEqual(capturer.values[:8], expected)
         # Task 'a' was or was not executed again, both cases are ok.
         self.assertIsSuperAndSubsequence(capturer.values[8:], [
             'b.t RUNNING',
@@ -1174,7 +1225,7 @@ class RetryTest(utils.EngineTestBase):
                     'c.t REVERTED(None)',
                     'b.t REVERTING',
                     'b.t REVERTED(None)']
-        self.assertItemsEqual(capturer.values[:4], expected)
+        self.assertCountEqual(capturer.values[:4], expected)
         expected = ['test2_retry.r RETRYING',
                     'b.t PENDING',
                     'c.t PENDING',
@@ -1186,7 +1237,7 @@ class RetryTest(utils.EngineTestBase):
                     'c.t RUNNING',
                     'a.t SUCCESS(5)',
                     'c.t SUCCESS(5)']
-        self.assertItemsEqual(expected, capturer.values[4:])
+        self.assertCountEqual(expected, capturer.values[4:])
         self.assertEqual(st.SUCCESS, engine.storage.get_flow_state())
 
 
@@ -1202,7 +1253,7 @@ class RetryParallelExecutionTest(utils.EngineTestBase):
             utils.ConditionalTask('task2')
         )
         engine = self._make_engine(flow)
-        engine.task_notifier.register('*', waiting_task.callback)
+        engine.atom_notifier.register('*', waiting_task.callback)
         engine.storage.inject({'y': 2})
         with utils.CaptureListener(engine, capture_flow=False) as capturer:
             engine.run()
@@ -1226,7 +1277,7 @@ class RetryParallelExecutionTest(utils.EngineTestBase):
                     'task2.t RUNNING',
                     'task2.t SUCCESS(None)',
                     'task1.t SUCCESS(5)']
-        self.assertItemsEqual(capturer.values, expected)
+        self.assertCountEqual(capturer.values, expected)
 
     def test_when_subflow_fails_revert_success_tasks(self):
         waiting_task = utils.WaitForOneFromTask('task2', 'task1',
@@ -1238,7 +1289,7 @@ class RetryParallelExecutionTest(utils.EngineTestBase):
                 utils.ConditionalTask('task3'))
         )
         engine = self._make_engine(flow)
-        engine.task_notifier.register('*', waiting_task.callback)
+        engine.atom_notifier.register('*', waiting_task.callback)
         engine.storage.inject({'y': 2})
         with utils.CaptureListener(engine, capture_flow=False) as capturer:
             engine.run()
@@ -1269,7 +1320,7 @@ class RetryParallelExecutionTest(utils.EngineTestBase):
                     'task2.t SUCCESS(5)',
                     'task3.t RUNNING',
                     'task3.t SUCCESS(None)']
-        self.assertItemsEqual(capturer.values, expected)
+        self.assertCountEqual(capturer.values, expected)
 
 
 class SerialEngineTest(RetryTest, test.TestCase):
@@ -1305,8 +1356,7 @@ class ParallelEngineWithEventletTest(RetryTest, test.TestCase):
     def _make_engine(self, flow, defer_reverts=None, flow_detail=None,
                      executor=None):
         if executor is None:
-            executor = futurist.GreenThreadPoolExecutor()
-            self.addCleanup(executor.shutdown)
+            executor = 'greenthreads'
         return taskflow.engines.load(flow,
                                      flow_detail=flow_detail,
                                      backend=self.backend,
